@@ -1,4 +1,4 @@
-// Copyright 2010 Google Inc.
+// Copyright 2010 The Kyua Authors.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -38,8 +38,7 @@
 #include <lutok/state.ipp>
 
 #include "engine/exceptions.hpp"
-#include "engine/runner.hpp"
-#include "engine/testers.hpp"
+#include "engine/scheduler.hpp"
 #include "model/metadata.hpp"
 #include "model/test_program.hpp"
 #include "utils/config/exceptions.hpp"
@@ -56,7 +55,7 @@
 namespace config = utils::config;
 namespace datetime = utils::datetime;
 namespace fs = utils::fs;
-namespace runner = engine::runner;
+namespace scheduler = engine::scheduler;
 
 using utils::none;
 using utils::optional;
@@ -179,9 +178,12 @@ public:
     ///     source_root_.
     /// \param user_config User configuration holding any test suite properties
     ///     to be passed to the list operation.
+    /// \param scheduler_handle The scheduler context to use for loading the
+    ///     test case lists.
     parser(const fs::path& source_root_, const fs::path& build_root_,
            const fs::path& relative_filename_,
-           const config::tree& user_config) :
+           const config::tree& user_config,
+           scheduler::scheduler_handle& scheduler_handle) :
         _source_root(source_root_), _build_root(build_root_),
         _relative_filename(relative_filename_)
     {
@@ -197,28 +199,32 @@ public:
         _state.set_global("current_kyuafile");
 
         *_state.new_userdata< const config::tree* >() = &user_config;
-        _state.push_cxx_closure(lua_include, 1);
+        *_state.new_userdata< scheduler::scheduler_handle* >() =
+            &scheduler_handle;
+        _state.push_cxx_closure(lua_include, 2);
         _state.set_global("include");
 
         _state.push_cxx_function(lua_test_suite);
         _state.set_global("test_suite");
 
         const std::set< std::string > interfaces =
-            engine::all_test_interfaces();
+            scheduler::registered_interface_names();
         for (std::set< std::string >::const_iterator iter = interfaces.begin();
              iter != interfaces.end(); ++iter) {
             const std::string& interface = *iter;
 
             _state.push_string(interface);
             *_state.new_userdata< const config::tree* >() = &user_config;
-            _state.push_cxx_closure(lua_generic_test_program, 2);
+            *_state.new_userdata< scheduler::scheduler_handle* >() =
+                &scheduler_handle;
+            _state.push_cxx_closure(lua_generic_test_program, 3);
             _state.set_global(interface + "_test_program");
         }
 
         _state.open_base();
         _state.open_string();
         _state.open_table();
-        fs::open_fs(_state);
+        fs::open_fs(_state, callback_current_kyuafile().branch_path());
     }
 
     /// Destructor.
@@ -260,14 +266,17 @@ public:
     /// \param raw_file Path to the file to include.
     /// \param user_config User configuration holding any test suite properties
     ///     to be passed to the list operation.
+    /// \param scheduler_handle Scheduler context to run test programs in.
     void
     callback_include(const fs::path& raw_file,
-                     const config::tree& user_config)
+                     const config::tree& user_config,
+                     scheduler::scheduler_handle& scheduler_handle)
     {
         const fs::path file = relativize(_relative_filename.branch_path(),
                                          raw_file);
         const model::test_programs_vector subtps =
-            parser(_source_root, _build_root, file, user_config).parse();
+            parser(_source_root, _build_root, file, user_config,
+                   scheduler_handle).parse();
 
         std::copy(subtps.begin(), subtps.end(),
                   std::back_inserter(_test_programs));
@@ -308,6 +317,7 @@ public:
     /// \param metadata Metadata variables passed to the test program.
     /// \param user_config User configuration holding any test suite properties
     ///     to be passed to the list operation.
+    /// \param scheduler_handle Scheduler context to run test programs in.
     ///
     /// \throw std::runtime_error If the test program definition is invalid or
     ///     if the test program does not exist.
@@ -316,7 +326,8 @@ public:
                           const fs::path& raw_path,
                           const std::string& test_suite_override,
                           const model::metadata& metadata,
-                          const config::tree& user_config)
+                          const config::tree& user_config,
+                          scheduler::scheduler_handle& scheduler_handle)
     {
         if (raw_path.is_absolute())
             throw std::runtime_error(F("Got unexpected absolute path for test "
@@ -334,12 +345,10 @@ public:
 
         const std::string test_suite = get_test_suite(test_suite_override);
 
-        const config::properties_map props = runner::generate_tester_config(
-            user_config, test_suite);
-
         _test_programs.push_back(model::test_program_ptr(
-            new runner::lazy_test_program(interface, path, _build_root,
-                                          test_suite, metadata, props)));
+            new scheduler::lazy_test_program(interface, path, _build_root,
+                                             test_suite, metadata, user_config,
+                                             scheduler_handle)));
     }
 
     /// Callback for the Kyuafile test_suite() function.
@@ -390,23 +399,6 @@ public:
 };
 
 
-/// Checks if the given interface name is valid.
-///
-/// \param interface The name of the interface to validate.
-///
-/// \throw std::runtime_error If the given interface is not supported.
-static void
-ensure_valid_interface(const std::string& interface)
-{
-    try {
-        (void)engine::tester_path(interface);
-    } catch (const engine::error& e) {
-        throw std::runtime_error(F("Unsupported test interface '%s'") %
-                                 interface);
-    }
-}
-
-
 /// Glue to invoke parser::callback_test_program() from Lua.
 ///
 /// This is a helper function for the various *_test_program() calls, as they
@@ -417,6 +409,7 @@ ensure_valid_interface(const std::string& interface)
 /// name.  The rest of the arguments are part of the test program metadata.
 /// \pre state(upvalue 1) String with the name of the interface.
 /// \pre state(upvalue 2) User configuration with the per-test suite settings.
+/// \pre state(upvalue 3) Scheduler context to run test programs in.
 ///
 /// \param state The Lua state that executed the function.
 ///
@@ -437,12 +430,19 @@ lua_generic_test_program(lutok::state& state)
     const config::tree* user_config = *state.to_userdata< const config::tree* >(
         state.upvalue_index(2));
 
+    if (!state.is_userdata(state.upvalue_index(3)))
+        throw std::runtime_error("Found corrupt state for test_program "
+                                 "function");
+    scheduler::scheduler_handle* scheduler_handle =
+        *state.to_userdata< scheduler::scheduler_handle* >(
+            state.upvalue_index(3));
+
     if (!state.is_table(-1))
         throw std::runtime_error(
             F("%s_test_program expects a table of properties as its single "
               "argument") % interface);
 
-    ensure_valid_interface(interface);
+    scheduler::ensure_valid_interface(interface);
 
     lutok::stack_cleaner cleaner(state);
 
@@ -477,11 +477,18 @@ lua_generic_test_program(lutok::state& state)
         const std::string property = state.to_string(-2);
 
         if (property != "name" && property != "test_suite") {
-            if (!state.is_number(-1) && !state.is_string(-1))
+            std::string value;
+            if (state.is_boolean(-1)) {
+                value = F("%s") % state.to_boolean(-1);
+            } else if (state.is_number(-1)) {
+                value = F("%s") % state.to_integer(-1);
+            } else if (state.is_string(-1)) {
+                value = state.to_string(-1);
+            } else {
                 throw std::runtime_error(
                     F("Metadata property '%s' in test program '%s' cannot be "
                       "converted to a string") % property % path);
-            const std::string value = state.to_string(-1);
+            }
 
             mdbuilder.set_string(property, value);
         }
@@ -490,7 +497,8 @@ lua_generic_test_program(lutok::state& state)
     }
 
     parser::get_from_state(state)->callback_test_program(
-        interface, path, test_suite, mdbuilder.build(), *user_config);
+        interface, path, test_suite, mdbuilder.build(), *user_config,
+        *scheduler_handle);
     return 0;
 }
 
@@ -514,6 +522,7 @@ lua_current_kyuafile(lutok::state& state)
 /// \param state The Lua state that executed the function.
 ///
 /// \pre state(upvalue 1) User configuration with the per-test suite settings.
+/// \pre state(upvalue 2) Scheduler context to run test programs in.
 ///
 /// \return Number of return values left on the Lua stack.
 static int
@@ -525,8 +534,15 @@ lua_include(lutok::state& state)
     const config::tree* user_config = *state.to_userdata< const config::tree* >(
         state.upvalue_index(1));
 
+    if (!state.is_userdata(state.upvalue_index(2)))
+        throw std::runtime_error("Found corrupt state for test_program "
+                                 "function");
+    scheduler::scheduler_handle* scheduler_handle =
+        *state.to_userdata< scheduler::scheduler_handle* >(
+            state.upvalue_index(2));
+
     parser::get_from_state(state)->callback_include(
-        fs::path(state.to_string(-1)), *user_config);
+        fs::path(state.to_string(-1)), *user_config, *scheduler_handle);
     return 0;
 }
 
@@ -616,6 +632,8 @@ engine::kyuafile::~kyuafile(void)
 ///     from which the Kyuafile is being read).
 /// \param user_config User configuration holding any test suite properties
 ///     to be passed to the list operation.
+/// \param scheduler_handle The scheduler context to use for loading the test
+///     case lists.
 ///
 /// \return High-level representation of the configuration file.
 ///
@@ -624,15 +642,25 @@ engine::kyuafile::~kyuafile(void)
 engine::kyuafile
 engine::kyuafile::load(const fs::path& file,
                        const optional< fs::path > user_build_root,
-                       const config::tree& user_config)
+                       const config::tree& user_config,
+                       scheduler::scheduler_handle& scheduler_handle)
 {
     const fs::path source_root_ = file.branch_path();
     const fs::path build_root_ = user_build_root ?
         user_build_root.get() : source_root_;
 
+    // test_program.absolute_path() uses the current work directory and that
+    // fails to resolve the correct path once we have used chdir to enter the
+    // test work directory.  To prevent this causing issues down the road,
+    // force the build root to be absolute so that absolute_path() does not
+    // need to rely on the current work directory.
+    const fs::path abs_build_root = build_root_.is_absolute() ?
+        build_root_ : build_root_.to_absolute();
+
     return kyuafile(source_root_, build_root_,
-                    parser(source_root_, build_root_,
-                           fs::path(file.leaf_name()), user_config).parse());
+                    parser(source_root_, abs_build_root,
+                           fs::path(file.leaf_name()), user_config,
+                           scheduler_handle).parse());
 }
 
 
