@@ -1,4 +1,4 @@
-// Copyright 2010 Google Inc.
+// Copyright 2010 The Kyua Authors.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -46,7 +46,6 @@ extern "C" {
 #endif
 #include <sys/wait.h>
 
-#include <dirent.h>
 #include <unistd.h>
 }
 
@@ -62,6 +61,7 @@ extern "C" {
 #include "utils/defs.hpp"
 #include "utils/env.hpp"
 #include "utils/format/macros.hpp"
+#include "utils/fs/directory.hpp"
 #include "utils/fs/exceptions.hpp"
 #include "utils/fs/path.hpp"
 #include "utils/logging/macros.hpp"
@@ -510,12 +510,35 @@ fs::mkdir_p(const fs::path& dir, const int mode)
 }
 
 
-/// Creates a temporary directory.
+/// Creates a temporary directory that is world readable/accessible.
 ///
 /// The temporary directory is created using mkdtemp(3) using the provided
 /// template.  This should be most likely used in conjunction with
 /// fs::auto_directory.
 ///
+/// The temporary directory is given read and execute permissions to everyone
+/// and thus should not be used to protect data that may be subject to snooping.
+/// This goes together with the assumption that this function is used to create
+/// temporary directories for test cases, and that those test cases may
+/// sometimes be executed as an unprivileged user.  In those cases, we need to
+/// support two different things:
+///
+/// - Allow the unprivileged code to write to files in the work directory by
+///   name (e.g. to write the results file, whose name is provided by the
+///   monitor code running as root).  This requires us to grant search
+///   permissions.
+///
+/// - Allow the test cases themselves to call getcwd(3) at any point.  At least
+///   on NetBSD 7.x, getcwd(3) requires both read and search permissions on all
+///   path components leading to the current directory.  This requires us to
+///   grant both read and search permissions.
+///
+/// TODO(jmmv): A cleaner way to support this would be for the test executor to
+/// create two work directory hierarchies directly rooted under TMPDIR: one for
+/// root and one for the unprivileged user.  However, that requires more
+/// bookkeeping for no real gain, because we are not really trying to protect
+/// the data within our temporary directories against attacks.
+///
 /// \param path_template The template for the temporary path, which is a
 ///     basename that is created within the TMPDIR.  Must contain the XXXXXX
 ///     pattern, which is atomically replaced by a random unique string.
@@ -524,7 +547,7 @@ fs::mkdir_p(const fs::path& dir, const int mode)
 ///
 /// \throw fs::system_error If the call to mkdtemp(3) fails.
 fs::path
-fs::mkdtemp(const std::string& path_template)
+fs::mkdtemp_public(const std::string& path_template)
 {
     PRE(path_template.find("XXXXXX") != std::string::npos);
 
@@ -539,7 +562,27 @@ fs::mkdtemp(const std::string& path_template)
                                  "template %s") % full_template,
                                original_errno);
     }
-    return fs::path(buf.get());
+    const fs::path path(buf.get());
+
+    if (::chmod(path.c_str(), 0755) == -1) {
+        const int original_errno = errno;
+
+        try {
+            rmdir(path);
+        } catch (const fs::system_error& e) {
+            // This really should not fail.  We just created the directory and
+            // have not written anything to it so there is no reason for this to
+            // fail.  But better handle the failure just in case.
+            LW(F("Failed to delete just-created temporary directory %s")
+               % path);
+        }
+
+        throw fs::system_error(F("Failed to grant search permissions on "
+                                 "temporary directory %s") % path,
+                               original_errno);
+    }
+
+    return path;
 }
 
 
@@ -574,6 +617,20 @@ fs::mkstemp(const std::string& path_template)
 }
 
 
+/// Mounts a temporary file system with unlimited size.
+///
+/// \param in_mount_point The path on which the file system will be mounted.
+///
+/// \throw fs::system_error If the attempt to mount process fails.
+/// \throw fs::unsupported_operation_error If the code does not know how to
+///     mount a temporary file system in the current operating system.
+void
+fs::mount_tmpfs(const fs::path& in_mount_point)
+{
+    mount_tmpfs(in_mount_point, units::bytes());
+}
+
+
 /// Mounts a temporary file system.
 ///
 /// \param in_mount_point The path on which the file system will be mounted.
@@ -636,34 +693,23 @@ retry:
 void
 fs::rm_r(const fs::path& directory)
 {
-    DIR* dirp = ::opendir(directory.c_str());
-    if (dirp == NULL) {
-        const int original_errno = errno;
-        throw fs::system_error(F("Failed to open directory %s") %
-                               directory.str(), original_errno);
-    }
-    try {
-        ::dirent* dp;
-        while ((dp = ::readdir(dirp)) != NULL) {
-            const std::string name = dp->d_name;
-            if (name == "." || name == "..")
-                continue;
+    const fs::directory dir(directory);
 
-            const fs::path entry = directory / dp->d_name;
+    for (fs::directory::const_iterator iter = dir.begin(); iter != dir.end();
+         ++iter) {
+        if (iter->name == "." || iter->name == "..")
+            continue;
 
-            if (fs::is_directory(entry)) {
-                LD(F("Descending into %s") % entry);
-                fs::rm_r(entry);
-            } else {
-                LD(F("Removing file %s") % entry);
-                fs::unlink(entry);
-            }
+        const fs::path entry = directory / iter->name;
+
+        if (fs::is_directory(entry)) {
+            LD(F("Descending into %s") % entry);
+            fs::rm_r(entry);
+        } else {
+            LD(F("Removing file %s") % entry);
+            fs::unlink(entry);
         }
-    } catch (...) {
-        ::closedir(dirp);
-        throw;
     }
-    ::closedir(dirp);
 
     LD(F("Removing empty directory %s") % directory);
     fs::rmdir(directory);
@@ -686,6 +732,28 @@ fs::rmdir(const path& file)
 }
 
 
+/// Obtains all the entries in a directory.
+///
+/// \param path The directory to scan.
+///
+/// \return The set of all directory entries in the given directory.
+///
+/// \throw fs::system_error If reading the directory fails for any reason.
+std::set< fs::directory_entry >
+fs::scan_directory(const fs::path& path)
+{
+    std::set< fs::directory_entry > contents;
+
+    fs::directory dir(path);
+    for (fs::directory::const_iterator iter = dir.begin(); iter != dir.end();
+         ++iter) {
+        contents.insert(*iter);
+    }
+
+    return contents;
+}
+
+
 /// Removes a file.
 ///
 /// \param file The file to remove.
